#include "Psf.h"
#include <limits.h>
#include <math.h>

#define MIN_ 100000

typedef struct {
  byte numChannels;
  ulong width;
  ulong height;
  ulong numEntries;
  PsfEntry *entries;
  double ***inChannels;
  double ***outChannels;
} Pass_;

static void blur_(Parallel *par);


void psf_new(Psf **p_this, DoubleImage *image) {
  ulong e;
  ulong i;
  ulong i_center;
  ulong in_height;
  ulong in_width;
  ulong j;
  ulong j_center;
  ulong n;
  long max_i;
  long max_j;
  long min_i;
  long min_j;
  double scale;
  double sum;
  double sum_v;
  double sum_v_i;
  double sum_v_j;
  double v;
  double **pixels;
  PsfEntry *entry;
  PsfEntry *entries;
  Psf *this;

  this = newo(p_this);

  if (!double_image_is_grayscale(image)) {
    throw("Point spread function image must be grayscale%s",
        double_image_has_alpha(image) ? " without transparency" : "");
  }

  in_width = image->width;
  in_height = image->height;
  pixels = image->y;

  n = 0;
  sum = 0;

  for (i = 0; i < in_width; i++) {
    sum += pixels[i][0];
    sum += pixels[i][in_height - 1];
    n += 2;
  }

  for (j = 1; j < in_height - 1; j++) {
    sum += pixels[0][j];
    sum += pixels[in_width - 1][j];
    n += 2;
  }

  if (sum / n > 0.5) {
    for (i = 0; i < in_width; i++) {
      for (j = 0; j < in_height; j++) {
        pixels[i][j] = 1 - pixels[i][j];
      }
    }
  }

  sum_v = 0;
  sum_v_i = 0;
  sum_v_j = 0;

  min_i = LONG_MAX;
  max_i = LONG_MIN;
  min_j = LONG_MAX;
  max_j = LONG_MIN;

  for (i = 0, n = 0, sum = 0; i < in_width; i++) {
    for (j = 0; j < in_height; j++) {
      v = pixels[i][j];

      if (fabs(v) > 1e-4) {
        n++;
        sum_v += v;
        sum_v_i += v * i;
        sum_v_j += v * j;
        min_i = minimum(min_i, (long)i);
        max_i = maximum(max_i, (long)i);
        min_j = minimum(min_j, (long)j);
        max_j = maximum(max_j, (long)j);
      }
    }
  }

  if (n == 0) {
    throw("No non-background data found in point spread function image");
  }

  this->numEntries = n;

  new(&this->_entries, n);
  entries = this->_entries;

  scale = 1 / sum_v;

  i_center = (ulong)round(sum_v_i * scale);
  j_center = (ulong)round(sum_v_j * scale);

  for (i = min_i, e = 0, sum = 0; i <= max_i; i++) {
    for (j = min_j; j <= max_j; j++) {
      v = pixels[i][j];

      if (fabs(v) > 1e-4) {
        entry = &entries[e];
        entry->i = (long)i - (long)i_center;
        entry->j = (long)j - (long)j_center;
        sum += (entry->w = v * scale);
        e++;
      }
    }
  }

  if (e != n) {
    throw("impossible");
  }

  if (fabs(sum - 1) > 1e-8) {
    throw("Impossible: %.16g (%.16g)", sum, sum - 1);
  }
}

void psf_new_from_filename(Psf **p_this, char *filename) {
  ByteImage *_bi = null;
  DoubleImage *_di = null;

  byte_image_new_load_file(&_bi, filename);
  double_image_new_from_byte(&_di, _bi);
  byte_image_delete(&_bi);

  psf_new(p_this, _di);

  double_image_delete(&_di);
}

void psf_delete(Psf **p_this) {
  Psf *this;

  if ((this = *p_this) == null) {
    return;
  }

  deletev(&this->_entries);

  deleteo(p_this);
}

void psf_print(Psf *this, Printer *p) {
  ulong u;
  PsfEntry *entry;

  sprint(p, "Psf");
  rprint(p, "Number of entries: %lu", this->numEntries);
  sprint(p, "Entries");

  for (u = 0; u < this->numEntries; u++) {
    entry = &this->_entries[u];
    rprint(p,
        "[%lu] i = %ld, j = %ld, w = %.16g",
        u,
        entry->i,
        entry->j,
        entry->w);
  }

  fprint(p);
  fprint(p);
}

DoubleImage *double_image_new_load_psf(DoubleImage **p_this, char *filename) {
  return null;
}

DoubleImage *double_image_new_blur_psf(
    DoubleImage **p_this,
    DoubleImage *other,
    Psf *psf) {
  byte num_channels;
  ulong height;
  ulong width;
  ulong num_entries;
  double ***in_channels;
  double ***out_channels;
  PsfEntry *entries;
  Pass_ pass;
  DoubleImage *this;

  width = other->width;
  height = other->height;
  num_channels = other->numChannels;
  in_channels = other->_channels;

  num_entries = psf->numEntries;
  entries = psf->_entries;

  double_image_new(p_this, width, height, num_channels);
  this = *p_this;

  out_channels = this->_channels;

  pass.numChannels = num_channels;
  pass.width = width;
  pass.height = height;
  pass.numEntries = num_entries;
  pass.entries = entries;
  pass.inChannels = in_channels;
  pass.outChannels = out_channels;
  parallel_for(width, MIN_ / psf->numEntries, blur_, &pass);

  return this;
}


static void blur_(Parallel *par) {
  byte c;
  ulong e;
  ulong i;
  ulong j;
  long in_i;
  long in_j;
  double sum;
  double w;
  double **in_channel;
  double **out_channel;
  PsfEntry *entry;
  Pass_ *pass;

  pass = par->x;

  for (c = 0; c < pass->numChannels; c++) {
    in_channel = pass->inChannels[c];
    out_channel = pass->outChannels[c];

    for (i = par->start; i < par->end; i++) {
      for (j = 0; j < pass->height; j++) {
        sum = 0;

        for (e = 0; e < pass->numEntries; e++) {
          entry = &pass->entries[e];

          in_i = bound((long)i - entry->i, 0, pass->width - 1);
          in_j = bound((long)j - entry->j, 0, pass->height - 1);

          w = entry->w;
          sum += w * in_channel[in_i][in_j];
        }

        out_channel[i][j] = sum;
      }
    }
  }
}
