#include "unblur.h"
#include <float.h>
#include <math.h>

#define MIN_ 100000

typedef struct {
  bool useWeights;
  byte numChannels;
  ulong width;
  ulong height;
  ulong numEntries;
  PsfEntry *entries;
  double ***inChannels;
  double ***outChannels;
} Pass_;

static DoubleImage *back_propagate_(
    DoubleImage **p_this,
    DoubleImage *other,
    Psf *psf,
    bool use_weights);

static void back_prop_(Parallel *par);


DoubleImage *double_image_new_unblur(
    DoubleImage **p_this,
    DoubleImage *other,
    Psf *psf,
    DoubleImage *approx,
    double tolerance,
    ulong max_iterations,
    Logger *l) {
  bool recalc;
  ulong iteration;
  double last_mean_error;
  double mean_error;
  DoubleImage *_delta = null;
  DoubleImage *_error = null;
  DoubleImage *_estimate = null;
  DoubleImage *_old_estimate = null;
  DoubleImage *_reblurred = null;

  rlog(l, "Tolerance in mean absolute error: %.16g%%.", 100 * tolerance);
  rlog(l, "Maximum number of iterations: %lu.", max_iterations);

  if (approx == null) {
    slog(l, "Calculating error in original input image");
    slog(l, "Blurring original input image");
    double_image_new_blur_psf(&_reblurred, other, psf);
    flog(l);
    slog(l, "Computing difference between blurred and original images");
    double_image_new_subtract(&_error, other, _reblurred);
    mean_error = double_image_mean_abs_rgb(_error);
    double_image_delete(&_error);
    double_image_delete(&_reblurred);
    rlog(l, "Mean absolute error of input image: %.16g%%.", 100 * mean_error);
    flog(l);
    flog(l);
    slog(l, "Creating initial estimate");
    double_image_new_back_propagate_psf(&_estimate, other, psf, l);
    flog(l);
  } else {
    slog(l, "Copying initial estimate");
    double_image_new_copy(&_estimate, approx);
    flog(l);
    mean_error = DBL_MAX;
  }

  recalc = true;

  for (iteration = 0; iteration < max_iterations; iteration++) {
    slog(l, "Iteration %lu (max %lu)", iteration + 1, max_iterations);

    slog(l, "Reblurring estimate");
    double_image_new_blur_psf(&_reblurred, _estimate, psf);
    flog(l);

    slog(l, "Calculating error in blurred space");
    double_image_new_subtract(&_error, other, _reblurred);
    double_image_delete(&_reblurred);
    flog(l);

    last_mean_error = mean_error;
    mean_error = double_image_mean_abs_rgb(_error);
    rlog(l, "Mean absolute error: %.16g%%.", 100 * mean_error);

    if (mean_error <= tolerance) {
      double_image_delete(&_error);
      rlog(l, "Mean absolute error is within tolerance: we're finished.");
      flog(l);
      recalc = false;
      break;
    }

    if (mean_error > last_mean_error) {
      if (approx == null && iteration == 0) {
        rlog(l,
            "Mean absolute error greater than that of the input image: "
            "reverting.");
        double_image_delete(&_estimate);
        double_image_delete(&_error);
        slog(l, "Reverting to input image");
        slog(l, "Copying input image");
        double_image_new_copy(&_estimate, other);
        flog(l);
        slog(l, "Blurring original input image");
        double_image_new_blur_psf(&_reblurred, other, psf);
        flog(l);
        slog(l,
            "Computing difference between blurred and original images");
        double_image_new_subtract(&_error, other, _reblurred);
        mean_error = double_image_mean_abs_rgb(_error);
        double_image_delete(&_reblurred);
        flog(l);
        flog(l);
      } else {
        double_image_delete(&_error);
        rlog(l,
            "WARNING: Mean absolute error increased from %.16g%% to %.16g%%: "
            "reverting and aborting.",
            100 * last_mean_error,
            100 * mean_error);
        double_image_delete(&_estimate);

        if (_old_estimate == null) {
          double_image_new_copy(&_estimate, other);
        } else {
          donate(&_old_estimate, &_estimate);
        }

        mean_error = last_mean_error;
        flog(l);
        recalc = false;
        break;
      }
    }

    slog(l, "Back-propagating error to estimate space");
    double_image_new_back_propagate_psf(&_delta, _error, psf, l);
    double_image_delete(&_error);
    flog(l);

    slog(l, "Adding back-propagated delta to our estimate");
    double_image_delete(&_old_estimate);
    donate(&_estimate, &_old_estimate);
    double_image_new_add(&_estimate, _old_estimate, _delta);
    double_image_delete(&_delta);
    flog(l);

    flog(l);
  }

  if (recalc) {
    slog(l, "Reblurring estimate");
    double_image_new_blur_psf(&_reblurred, _estimate, psf);
    flog(l);

    slog(l, "Calculating error in blurred space");
    double_image_new_subtract(&_error, other, _reblurred);
    double_image_delete(&_reblurred);
    flog(l);

    mean_error = double_image_mean_abs_rgb(_error);
    double_image_delete(&_error);
  }

  double_image_delete(&_old_estimate);
  rlog(l, "Mean absolute error of final estimate: %.16g%%.", 100 * mean_error);
  donate(&_estimate, p_this);

  return 0;
}

DoubleImage *double_image_new_back_propagate_psf(
    DoubleImage **p_this,
    DoubleImage *other,
    Psf *psf,
    Logger *l) {
  byte c;
  byte num_channels;
  ulong height;
  ulong i;
  ulong j;
  ulong width;
  double in_v;
  double multiplier_v;
  double reblurred_v;
  double **back_multiplier_channel;
  double **in_channel;
  double **initial_channel;
  double **multiplier_channel;
  double **out_channel;
  double **reblurred_channel;
  double ***back_multiplier_channels;
  double ***in_channels;
  double ***initial_channels;
  double ***multiplier_channels;
  double ***out_channels;
  double ***reblurred_channels;
  DoubleImage *_back_multiplier = null;
  DoubleImage *_initial = null;
  DoubleImage *_multiplier = null;
  DoubleImage *_reblurred = null;
  DoubleImage *this;

  width = other->width;
  height = other->height;
  num_channels = other->numChannels;
  in_channels = other->_channels;

  slog(l, "Creating initial estimate by back propagation of input image");
  back_propagate_(&_initial, other, psf, true);
  initial_channels = _initial->_channels;
  flog(l);

  slog(l, "Reblurring initial estimate");
  double_image_new_blur_psf(&_reblurred, _initial, psf);
  reblurred_channels = _reblurred->_channels;
  flog(l);

  slog(l, "Calculating multipliers in blurred space");
  double_image_new(&_multiplier, width, height, num_channels);
  multiplier_channels = _multiplier->_channels;

  for (c = 0; c < num_channels; c++) {
    in_channel = in_channels[c];
    reblurred_channel = reblurred_channels[c];
    multiplier_channel = multiplier_channels[c];

    for (i = 0; i < width; i++) {
      for (j = 0; j < height; j++) {
        in_v = in_channel[i][j];
        reblurred_v = reblurred_channel[i][j];

        if (reblurred_v == 0) {
          multiplier_v = 1;
        } else if (reblurred_v * in_v <= 0) {
          multiplier_v = 0;
        } else {
          multiplier_v = minimum(in_v / reblurred_v, 1);
        }

        multiplier_channel[i][j] = multiplier_v;
      }
    }
  }

  double_image_delete(&_reblurred);
  flog(l);

  slog(l, "Creating multipliers in estimate space by back-propagation");
  back_propagate_(&_back_multiplier, _multiplier, psf, false);
  back_multiplier_channels = _back_multiplier->_channels;
  double_image_delete(&_multiplier);
  flog(l);

  slog(l, "Creating final estimate");

  double_image_new(p_this, width, height, num_channels);
  this = *p_this;

  out_channels = this->_channels;

  for (c = 0; c < num_channels; c++) {
    initial_channel = initial_channels[c];
    out_channel = out_channels[c];
    back_multiplier_channel = back_multiplier_channels[c];

    for (i = 0; i < width; i++) {
      for (j = 0; j < height; j++) {
        out_channel[i][j]
          = back_multiplier_channel[i][j] * initial_channel[i][j];
      }
    }
  }

  double_image_delete(&_back_multiplier);
  double_image_delete(&_initial);
  flog(l);

  return this;
}


static DoubleImage *back_propagate_(
    DoubleImage **p_this,
    DoubleImage *other,
    Psf *psf,
    bool use_weights) {
  byte num_channels;
  ulong height;
  ulong min_parallel;
  ulong num_entries;
  ulong width;
  double ***in_channels;
  double ***out_channels;
  PsfEntry *entries;
  Pass_ pass;
  DoubleImage *this;

  width = other->width;
  height = other->height;
  num_channels = other->numChannels;
  in_channels = other->_channels;

  num_entries = psf->numEntries;
  entries = psf->_entries;

  double_image_new(p_this, width, height, num_channels);
  this = *p_this;

  out_channels = this->_channels;

  pass.numChannels = num_channels;
  pass.width = width;
  pass.height = height;
  pass.numEntries = num_entries;
  pass.entries = entries;
  pass.useWeights = use_weights;
  pass.inChannels = in_channels;
  pass.outChannels = out_channels;
  min_parallel = MIN_ / (num_channels * height * num_entries);
  parallel_for(width, min_parallel, back_prop_, &pass);

  return this;
}

static void back_prop_(Parallel *par) {
  bool single_median;
  byte c;
  ulong e;
  ulong i;
  ulong in_i;
  ulong in_j;
  ulong j;
  double estimate;
  double *_estimates = null;
  double *median_1;
  double *median_2;
  double **in_channel;
  double **out_channel;
  PsfEntry *entry;
  Pass_ *pass;

  pass = par->x;

  new(&_estimates, pass->numEntries);

  if (pass->numEntries % 2 == 1) {
    single_median = true;
    median_1 = &_estimates[(pass->numEntries - 1) / 2];
  } else {
    single_median = false;
    median_1 = &_estimates[pass->numEntries / 2 - 1];
    median_2 = &_estimates[pass->numEntries / 2];
  }

  for (c = 0; c < pass->numChannels; c++) {
    in_channel = pass->inChannels[c];
    out_channel = pass->outChannels[c];

    for (i = par->start; i < par->end; i++) {
      for (j = 0; j < pass->height; j++) {
        for (e = 0; e < pass->numEntries; e++) {
          entry = &pass->entries[e];
          in_i = bound((long)i + entry->i, 0, pass->width - 1);
          in_j = bound((long)j + entry->j, 0, pass->height - 1);;

          estimate = in_channel[in_i][in_j];

          if (pass->useWeights) {
            estimate /= entry->w;
          }

          _estimates[e] = estimate;
        }

        qsortx(_estimates, pass->numEntries, double_compare, true, null);

        if (single_median) {
          estimate = *median_1;
        } else {
          estimate = 0.5 * (*median_1 + *median_2);
        }

        out_channel[i][j] = estimate;
      }
    }
  }

  deletev(&_estimates);
}
