#include "unblur_image.h"

#define DEFAULT_TOLERANCE_ 1
#define DEFAULT_MAX_ITERATIONS_ 10

int main(int argc, char *argv[]) {
  byte quality;
  char *approx_filename;
  char *input_filename;
  char *output_filename;
  char *psf_filename;
  double tolerance;
  ByteImage *_approx = null;
  ByteImage *_in = null;
  ByteImage *_out = null;
  DoubleImage *_dapprox = null;
  DoubleImage *_din = null;
  DoubleImage *_dout = null;
  Psf *_psf = null;
  Logger *_l = null;

  args_description("Deblur an image using a point spread function image");
  args_string("input-filename", "Input image to be deblurred");
  args_string("psf-filename", "Post spread function image filename");
  args_string("output-filename", "Deblurred output image to be created");
  args_flag("quiet", 'q', "Quiet mode");
  args_optional_real(
      "tolerance",
      't',
      DEFAULT_TOLERANCE_,
      "Mean absolute error tolerance: if it drops below this number of "
      "intensity units, across all pixels of all channels, then we're done "
      "(max\f=\f255)");
  args_optional_integer(
      "max-iterations",
      'i',
      DEFAULT_MAX_ITERATIONS_,
      "Maximum number of iterations");
  args_optional_string(
      "initial-approx-filename",
      'a',
      ARGS_STRING_NO_DEFAULT,
      "Initial approximation filename");
  args_flag("auto-scale", 's', "Auto-scale the output intensity range");
  args_flag("png", 'p', "Save the output file in PNG format");
  args_flag("jpeg", 'j', "Save the output file in JPEG format");
  args_optional_integer(
      "jpeg-quality",
      'J',
      jpeg_file_get_default_quality(),
      "The JPEG quality to use for the output JPEG file, between 0 and 100");
  args_parse(argc, argv);

  logger_new(&_l, args_set("quiet") ? null : stdout);

  input_filename = args_get_string("input-filename");
  psf_filename = args_get_string("psf-filename");
  approx_filename = args_get_string("initial-approx-filename");
  output_filename = args_get_string("output-filename");
  tolerance = args_get_bounded_positive_real("tolerance", 255);

  if (tolerance < 0.25) {
    rlog(_l,
        "WARNING: Tolerance of less than 1/4 unit of intensity is "
        "unrealistic.");
  }

  slog(_l, "Loading blurred input image from '%s'", input_filename);
  byte_image_new_load_file(&_in, input_filename);
  double_image_new_from_byte(&_din, _in);
  byte_image_delete(&_in);
  flog(_l);

  slog(_l, "Loading point spread function image from '%s'", psf_filename);
  psf_new_from_filename(&_psf, psf_filename);
  flog(_l);

  if (approx_filename != null) {
    slog(_l, "Loading initial approximation image from '%s'", approx_filename);
    byte_image_new_load_file(&_approx, approx_filename);
    double_image_new_from_byte(&_dapprox, _approx);
    byte_image_delete(&_approx);
    flog(_l);
  }

  double_image_new_unblur(
      &_dout,
      _din,
      _psf,
      _dapprox,
      tolerance / 255,
      args_get_positive_integer("max-iterations"),
      _l);
  psf_delete(&_psf);
  double_image_delete(&_dapprox);
  double_image_delete(&_din);

  if (args_set("auto-scale")) {
    double_image_auto_scale(_dout);
  }

  byte_image_new_from_double(&_out, _dout);
  double_image_delete(&_dout);

  quality = args_get_bounded_integer("jpeg-quality", 0, 100);

  slog(_l, "Writing unblurred result to '%s'", output_filename);

  if (args_get_flag_veto("png", "jpeg")) {
    byte_image_save_png(_out, output_filename);
  } else if (args_get_flag("jpeg")) {
    byte_image_save_jpeg(_out, output_filename, quality);
  } else {
    byte_image_save_quality(_out, output_filename, quality);
  }

  byte_image_delete(&_out);
  flog(_l);

  logger_delete(&_l);

  return 0;
}
